@@ -16,13 +16,14 @@
 """
 Routines for formatting opcodes.
 """
-from typing import Optional, Tuple
+from typing import List, Optional, Tuple
 
+from xdis.instruction import Instruction
 from xdis.opcodes.format.basic import format_IS_OP, format_RAISE_VARARGS_older
 
 
 def extended_format_binary_op(
-    opc, instructions: list, fmt_str: str
+    opc, instructions: List[Instruction], fmt_str: str
 ) -> Tuple[str, Optional[int]]:
     """
     General routine for formatting binary operations.
@@ -40,33 +41,40 @@ def extended_format_binary_op(
     i = skip_cache(instructions, 1)
     stack_inst1 = instructions[i]
     arg1 = None
-    if stack_inst1.tos_str is not None:
-        arg1 = stack_inst1.tos_str
-    if arg1 is not None or stack_inst1.opcode in opc.operator_set:
-        if arg1 is None:
-            arg1 = stack_inst1.argrepr
-        arg1_start_offset = stack_inst1.start_offset
-        if arg1_start_offset is not None:
-            for i in range(1, len(instructions)):
-                if instructions[i].offset == arg1_start_offset:
-                    break
-        j = skip_cache(instructions, i + 1)
-        stack_inst2 = instructions[j]
-        if (
-            stack_inst1.opcode in opc.operator_set
-            and stack_inst2.opcode in opc.operator_set
-        ):
-            arg2 = get_instruction_arg(stack_inst2, stack_inst2.argrepr)
-            start_offset = stack_inst2.start_offset
-            return fmt_str % (arg2, arg1), start_offset
-        elif stack_inst2.start_offset is not None:
-            start_offset = stack_inst2.start_offset
-            arg2 = get_instruction_arg(stack_inst2, stack_inst2.argrepr)
-            if arg2 == "":
-                arg2 = "..."
-            return fmt_str % (arg2, arg1), start_offset
-        else:
-            return fmt_str % ("...", arg1), None
+
+    # If stack_inst1 is a jump target, then its predecessor stack_inst2
+    # is possibly one of two values.
+    if not stack_inst1.is_jump_target:
+        if stack_inst1.tos_str is not None:
+            arg1 = stack_inst1.tos_str
+        if arg1 is not None or stack_inst1.opcode in opc.operator_set:
+            if arg1 is None:
+                arg1 = stack_inst1.argrepr
+            arg1_start_offset = stack_inst1.start_offset
+            if arg1_start_offset is not None:
+                i = get_instruction_index_from_offset(
+                    arg1_start_offset, instructions, 1
+                )
+                if i is None:
+                    return "", None
+            j = skip_cache(instructions, i + 1)
+            stack_inst2 = instructions[j]
+            if (
+                stack_inst1.opcode in opc.operator_set
+                and stack_inst2.opcode in opc.operator_set
+                and not stack_inst2.is_jump_target
+            ):
+                arg2 = get_instruction_arg(stack_inst2, stack_inst2.argrepr)
+                start_offset = stack_inst2.start_offset
+                return fmt_str % (arg2, arg1), start_offset
+            elif stack_inst2.start_offset is not None:
+                start_offset = stack_inst2.start_offset
+                arg2 = get_instruction_arg(stack_inst2, stack_inst2.argrepr)
+                if arg2 == "":
+                    arg2 = "..."
+                return fmt_str % (arg2, arg1), start_offset
+            else:
+                return fmt_str % ("...", arg1), None
     return "", None
 
 
@@ -89,9 +97,9 @@ def extended_format_infix_binary_op(
             arg1 = f"({arg1})"
         arg1_start_offset = instructions[1].start_offset
         if arg1_start_offset is not None:
-            for i in range(1, len(instructions)):
-                if instructions[i].offset == arg1_start_offset:
-                    break
+            i = get_instruction_index_from_offset(arg1_start_offset, instructions, 1)
+            if i is None:
+                return "", None
         j = i + 1
         # 3.11+ has CACHE instructions
         while instructions[j].opname == "CACHE":
@@ -124,16 +132,22 @@ def extended_format_infix_binary_op(
     return "", None
 
 
-def extended_format_store_op(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_store_op(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     inst = instructions[0]
+
+    # If the store instruction is a jump target, then
+    # the previous instruction is ambiguous. Here, things
+    # are more complicated, so let's not try to figure this out.
+    # This kind of things is best left for a decompiler.
+    if inst.is_jump_target:
+        return "", None
+
     prev_inst = instructions[1]
     start_offset = prev_inst.offset
-    if prev_inst.opname == "IMPORT_NAME":
-        return f"{inst.argval} = import_module({inst.argval})", start_offset
-    elif prev_inst.opname == "IMPORT_FROM":
-        return f"{inst.argval} = import_module({prev_inst.argval})", start_offset
-    elif prev_inst.opcode in opc.operator_set:
-        if prev_inst.opname == "LOAD_CONST":
+    if prev_inst.opcode in opc.operator_set:
+        if prev_inst.opcode in opc.nullaryloadop:
             argval = safe_repr(prev_inst.argval)
         elif (
             prev_inst.opcode in opc.VARGS_OPS | opc.NARGS_OPS
@@ -149,7 +163,7 @@ def extended_format_store_op(opc, instructions: list) -> Tuple[str, Optional[int
         argval = get_instruction_arg(prev_inst, argval)
         start_offset = prev_inst.start_offset
         if prev_inst.opname.startswith("INPLACE_"):
-            # Inplace operators show their own assign
+            # Inplace operators have their own assign routine.
             return argval, start_offset
         return f"{inst.argval} = {argval}", start_offset
 
@@ -182,28 +196,31 @@ def extended_format_ternary_op(
             arg1 = stack_inst1.argrepr
         arg1_start_offset = stack_inst1.start_offset
         if arg1_start_offset is not None:
-            for i in range(1, len(instructions)):
-                if instructions[i].offset == arg1_start_offset:
-                    break
+            i = get_instruction_index_from_offset(arg1_start_offset, instructions, 1)
+            if i is None:
+                return "", None
         j = skip_cache(instructions, i + 1)
         stack_inst2 = instructions[j]
         if (
             stack_inst1.opcode in opc.operator_set
             and stack_inst2.opcode in opc.operator_set
+            and not stack_inst2.is_jump_target
         ):
             arg2 = get_instruction_arg(stack_inst2, stack_inst2.argrepr)
             k = skip_cache(instructions, j + 1)
-            stack_inst3 = instructions[k]
-            if stack_inst3.opcode in opc.operator_set:
-                start_offset = stack_inst3.start_offset
+            stack_inst3 = instructions[k + 1]
+            start_offset = stack_inst3.start_offset
+            if (
+                stack_inst3.opcode in opc.operator_set
+                and not stack_inst3.is_jump_target
+            ):
                 arg3 = get_instruction_arg(stack_inst3, stack_inst3.argrepr)
                 return fmt_str % (arg2, arg1, arg3), start_offset
             else:
-                start_offset = stack_inst2.start_offset
                 arg3 = "..."
                 return fmt_str % (arg2, arg1, arg3), start_offset
 
-        elif stack_inst2.start_offset is not None:
+        elif stack_inst2.start_offset is not None and not stack_inst2.is_jump_target:
             start_offset = stack_inst2.start_offset
             arg2 = get_instruction_arg(stack_inst2, stack_inst2.argrepr)
             if arg2 == "":
@@ -215,7 +232,9 @@ def extended_format_ternary_op(
     return "", None
 
 
-def extended_format_STORE_SUBSCR(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_STORE_SUBSCR(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_ternary_op(
         opc,
         instructions,
@@ -228,63 +247,90 @@ def extended_format_unary_op(
 ) -> Tuple[str, Optional[int]]:
     stack_arg = instructions[1]
     start_offset = instructions[1].start_offset
-    if stack_arg.tos_str is not None:
+    if stack_arg.tos_str is not None and not stack_arg.is_jump_target:
         return fmt_str % stack_arg.tos_str, start_offset
     if stack_arg.opcode in opc.operator_set:
         return fmt_str % stack_arg.argrepr, start_offset
     return "", None
 
 
-def extended_format_ATTR(opc, instructions: list) -> Tuple[str, Optional[int]]:
-    if instructions[1].opcode in opc.NAME_OPS | opc.CONST_OPS | opc.LOCAL_OPS:
+def extended_format_ATTR(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
+    """
+    Handles both LOAD_ATTR and STORE_ATTR
+    """
+    instr1 = instructions[1]
+    if (
+        instr1.tos_str
+        or instr1.opcode in opc.NAME_OPS | opc.CONST_OPS | opc.LOCAL_OPS | opc.FREE_OPS
+    ):
+        base = get_instruction_arg(instr1)
+
         return (
-            f"{instructions[0].argrepr}.{instructions[1].argrepr}",
-            instructions[1].offset,
+            f"{base}.{instructions[0].argrepr}",
+            instr1.start_offset,
         )
     return "", None
 
 
-def extended_format_BINARY_ADD(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BINARY_ADD(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " + ")
 
 
-def extended_format_BINARY_AND(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BINARY_AND(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " & ")
 
 
 def extended_format_BINARY_FLOOR_DIVIDE(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " // ")
 
 
-def extended_format_BINARY_LSHIFT(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BINARY_LSHIFT(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " << ")
 
 
-def extended_format_BINARY_MODULO(opc, instructions: list) -> Tuple[str, Optional[int]]:
-    return extended_format_infix_binary_op(opc, instructions, " %% ")
+def extended_format_BINARY_MODULO(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
+    return extended_format_infix_binary_op(opc, instructions, " % ")
 
 
 def extended_format_BINARY_MULTIPLY(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " * ")
 
 
-def extended_format_BINARY_OR(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BINARY_OR(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " | ")
 
 
-def extended_format_BINARY_POWER(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BINARY_POWER(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " ** ")
 
 
-def extended_format_BINARY_RSHIFT(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BINARY_RSHIFT(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " >> ")
 
 
-def extended_format_BINARY_SUBSCR(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BINARY_SUBSCR(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_binary_op(
         opc,
         instructions,
@@ -293,47 +339,87 @@ def extended_format_BINARY_SUBSCR(opc, instructions: list) -> Tuple[str, Optiona
 
 
 def extended_format_BINARY_SUBTRACT(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " - ")
 
 
 def extended_format_BINARY_TRUE_DIVIDE(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " / ")
 
 
-def extended_format_BINARY_XOR(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BINARY_XOR(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " ^ ")
 
 
-def extended_format_BUILD_LIST(opc, instructions: list) -> Tuple[str, Optional[int]]:
-    if instructions[0].argval == 0:
-        # Degenerate case
-        return "[]", instructions[0].start_offset
+def extended_format_build_tuple_or_list(
+    opc, instructions: List[Instruction], left_delim: str, right_delim: str
+) -> Tuple[str, Optional[int]]:
+    arg_count = instructions[0].argval
+    is_tuple = left_delim == "("
+    if arg_count == 0:
+        # Note: caller generally handles this when the below isn't right.
+        return f"{left_delim}{right_delim}", instructions[0].offset
+    arglist, _, i = get_arglist(instructions, 0, arg_count)
+    if arglist is not None:
+        assert isinstance(i, int)
+        args_str = ", ".join(reversed(arglist))
+        if arg_count == 1 and is_tuple:
+            return f"{left_delim}{args_str},{right_delim}", instructions[i].start_offset
+        else:
+            return f"{left_delim}{args_str}{right_delim}", instructions[i].start_offset
     return "", None
 
 
-def extended_format_BUILD_MAP(opc, instructions: list) -> Tuple[str, Optional[int]]:
-    if instructions[0].argval == 0:
-        # Degenerate case
-        return "{}", instructions[0].start_offset
+def extended_format_BUILD_CONST_KEY_MAP(opc, instructions):
+    arg_count = instructions[0].argval
+    if arg_count == 0:
+        # Note: caller generally handles this when the below isn't right.
+        return "{}", instructions[0].offset
+    assert len(instructions) > 0
+    key_tuple = instructions[1]
+    key_values = key_tuple.argval
+    if key_tuple.opname == "LOAD_CONST" and isinstance(key_values, tuple):
+        arglist, _, i = get_arglist(instructions, 1, arg_count)
+        if arglist is not None:
+            assert isinstance(i, int)
+            assert len(arglist) == len(key_values)
+            arg_pairs = []
+            for i in range(len(arglist)):
+                arg_pairs.append(f"{key_values[i]}: {arglist[i]}")
+            args_str = ", ".join(arg_pairs)
+            return "{" + args_str + "}", instructions[i].start_offset
     return "", None
 
 
-def extended_format_BUILD_SET(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BUILD_LIST(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
+    return extended_format_build_tuple_or_list(opc, instructions, "[", "]")
+
+
+def extended_format_BUILD_SET(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     if instructions[0].argval == 0:
         # Degenerate case
         return "set()", instructions[0].start_offset
-    return "", None
+    return extended_format_build_tuple_or_list(opc, instructions, "{", "}")
 
 
-def extended_format_BUILD_SLICE(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BUILD_SLICE(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     argc = instructions[0].argval
+
     assert argc in (2, 3)
-    arglist, arg_count, i = get_arglist(instructions, 1, argc)
+    arglist, arg_count, i = get_arglist(instructions, 0, argc)
     if arg_count == 0:
+        assert isinstance(i, int)
         arglist = ["" if arg == "None" else arg for arg in arglist]
         return ":".join(reversed(arglist)), instructions[i].start_offset
 
@@ -343,21 +429,18 @@ def extended_format_BUILD_SLICE(opc, instructions: list) -> Tuple[str, Optional[
     return "", None
 
 
-def extended_format_BUILD_TUPLE(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_BUILD_TUPLE(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     arg_count = instructions[0].argval
     if arg_count == 0:
-        # Degenerate case
         return "tuple()", instructions[0].start_offset
-    arglist, _, i = get_arglist(instructions, 0, arg_count)
-    args_str = ", ".join(reversed(arglist))
-    if arg_count == 1:
-        return f"({args_str},)", instructions[i].start_offset
-    else:
-        return f"({args_str})", instructions[i].start_offset
-    return "", None
+    return extended_format_build_tuple_or_list(opc, instructions, "(", ")")
 
 
-def extended_format_COMPARE_OP(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_COMPARE_OP(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(
         opc,
         instructions,
@@ -380,7 +463,7 @@ def extended_format_CALL_FUNCTION(opc, instructions) -> Tuple[str, Optional[int]
 
     arglist, arg_count, i = get_arglist(instructions, 0, arg_count)
 
-    if arg_count != 0:
+    if arglist is None:
         return "", None
 
     assert i is not None
@@ -394,24 +477,48 @@ def extended_format_CALL_FUNCTION(opc, instructions) -> Tuple[str, Optional[int]
             arglist[0] = instructions[2].argval
 
         fn_name = fn_inst.tos_str if fn_inst.tos_str else fn_inst.argrepr
-        if opc.version_tuple >= (3, 6):
-            arglist.reverse()
+        arglist.reverse()
         s = f'{fn_name}({", ".join(arglist)})'
         return s, start_offset
     return "", None
 
 
-def extended_format_IMPORT_NAME(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_IMPORT_FROM(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
+    assert len(instructions) >= 2
+    i = 1
+    while instructions[i].opname == "STORE_NAME":
+        i = get_instruction_index_from_offset(
+            instructions[i].start_offset, instructions, 1
+        )
+        if i is None:
+            return "", None
+    module_name = get_instruction_arg(instructions[i])
+    if module_name.startswith("import_module("):
+        module_name = module_name[len("import_module(") : -1]
+    return (
+        f"from {module_name} import {instructions[0].argval}",
+        instructions[1].start_offset,
+    )
+
+
+def extended_format_IMPORT_NAME(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     inst = instructions[0]
-    start_offset = inst.start_offset
-    return f"import_module({inst.argval})", start_offset
+    return f"import_module({inst.argval})", inst.offset
 
 
-def extended_format_INPLACE_ADD(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_INPLACE_ADD(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " += ")
 
 
-def extended_format_INPLACE_AND(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_INPLACE_AND(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " &= ")
 
 
@@ -422,54 +529,62 @@ def extended_format_INPLACE_FLOOR_DIVIDE(
 
 
 def extended_format_INPLACE_LSHIFT(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " <<= ")
 
 
 def extended_format_INPLACE_MODULO(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " %%= ")
 
 
 def extended_format_INPLACE_MULTIPLY(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " *= ")
 
 
-def extended_format_INPLACE_OR(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_INPLACE_OR(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " |= ")
 
 
-def extended_format_INPLACE_POWER(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_INPLACE_POWER(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " **= ")
 
 
 def extended_format_INPLACE_TRUE_DIVIDE(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " /= ")
 
 
 def extended_format_INPLACE_RSHIFT(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " >>= ")
 
 
 def extended_format_INPLACE_SUBTRACT(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " -= ")
 
 
-def extended_format_INPLACE_XOR(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_INPLACE_XOR(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(opc, instructions, " ^= ")
 
 
-def extended_format_IS_OP(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_IS_OP(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_infix_binary_op(
         opc,
         instructions,
@@ -478,12 +593,14 @@ def extended_format_IS_OP(opc, instructions: list) -> Tuple[str, Optional[int]]:
 
 
 def extended_format_LOAD_BUILD_CLASS(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return "class", instructions[0].start_offset
 
 
-def extended_format_MAKE_FUNCTION_10_27(opc, instructions: list) -> Tuple[str, int]:
+def extended_format_MAKE_FUNCTION_10_27(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, int]:
     """
     instructions[0] should be a "MAKE_FUNCTION" or "MAKE_CLOSURE" instruction. TOS
     should have the function or closure name.
@@ -514,9 +631,38 @@ def extended_format_MAKE_FUNCTION_10_27(opc, instructions: list) -> Tuple[str, i
     return s, start_offset
 
 
+def extended_format_CALL_METHOD(opc, instructions) -> Tuple[str, Optional[int]]:
+    """call_method should be a "CALL_METHOD" instruction. Look in
+    `instructions` to see if we can find a method name.  If not we'll
+    return None.
+
+    """
+    # From opcode description: arg_count indicates the total number of
+    # positional and keyword arguments.
+
+    call_method_inst = instructions[0]
+    arg_count = call_method_inst.argval
+    s = ""
+
+    arglist, arg_count, first_arg = get_arglist(instructions, 0, arg_count)
+
+    if first_arg is None or first_arg >= len(instructions) - 1:
+        return "", None
+
+    fn_inst = instructions[first_arg + 1]
+    if fn_inst.opcode in opc.operator_set and arglist is not None:
+        start_offset = fn_inst.offset
+        if fn_inst.opname == "LOAD_METHOD":
+            fn_name = fn_inst.tos_str if fn_inst.tos_str else fn_inst.argrepr
+            arglist.reverse()
+            s = f'{fn_name}({", ".join(arglist)})'
+            return s, start_offset
+    return "", None
+
+
 def extended_format_RAISE_VARARGS_older(
-    opc, instructions: list
-) -> Tuple[Optional[str], int]:
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     raise_inst = instructions[0]
     assert raise_inst.opname == "RAISE_VARARGS"
     argc = raise_inst.argval
@@ -536,21 +682,27 @@ def extended_format_RAISE_VARARGS_older(
     return format_RAISE_VARARGS_older(raise_inst.argval), start_offset
 
 
-def extended_format_RETURN_VALUE(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_RETURN_VALUE(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_unary_op(opc, instructions, "return %s")
 
 
-def extended_format_UNARY_INVERT(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_UNARY_INVERT(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_unary_op(opc, instructions, "~(%s)")
 
 
 def extended_format_UNARY_NEGATIVE(
-    opc, instructions: list
+    opc, instructions: List[Instruction]
 ) -> Tuple[str, Optional[int]]:
     return extended_format_unary_op(opc, instructions, "-(%s)")
 
 
-def extended_format_UNARY_NOT(opc, instructions: list) -> Tuple[str, Optional[int]]:
+def extended_format_UNARY_NOT(
+    opc, instructions: List[Instruction]
+) -> Tuple[str, Optional[int]]:
     return extended_format_unary_op(opc, instructions, "not (%s)")
 
 
@@ -560,32 +712,37 @@ def extended_function_signature(code) -> str:
     """
     # FIXME: we can probably much better than this.
     # But this is a start.
-    return "..."
+    return "" if code.co_argcount == 0 else "..."
 
 
 def get_arglist(
-    instructions: list, i: int, arg_count: int
-) -> Tuple[list, int, Optional[int]]:
+    instructions: List[Instruction], i: int, arg_count: int
+) -> Tuple[Optional[list], int, Optional[int]]:
     """
     For a variable-length instruction like BUILD_TUPLE, or
-    a varlabie-name argument list, like CALL_FUNCTION
+    a variable-name argument list, like CALL_FUNCTION
     accumulate and find the beginning of the list and return:
     * argument list
+    * number of arguments parsed
     * the instruction index of the first instruction
 
     """
     arglist = []
     inst = None
     n = len(instructions) - 1
-    while arg_count > 0 and i < n:
+    to_do = arg_count
+    while to_do > 0 and i < n:
         i += 1
         inst = instructions[i]
-        arg_count -= 1
+        if inst.is_jump_target:
+            return None, -1, None
+
+        to_do -= 1
         arg = inst.tos_str if inst.tos_str else inst.argrepr
         if arg is not None:
             arglist.append(arg)
         elif not arg:
-            return arglist, arg_count, i
+            return arglist, arg_count - to_do, i
         else:
             arglist.append("???")
         if inst.is_jump_target:
@@ -606,15 +763,24 @@ def get_arglist(
                         start_offset = inst2.start_offset
 
         pass
-    return arglist, arg_count, i
+    return arglist, arg_count - to_do, i
 
 
-def get_instruction_arg(inst, argval=None) -> str:
+def get_instruction_arg(inst: Instruction, argval=None) -> str:
     argval = inst.argrepr if argval is None else argval
     return inst.tos_str if inst.tos_str is not None else argval
 
 
-def resolved_attrs(instructions: list) -> Tuple[str, int]:
+def get_instruction_index_from_offset(
+    target_offset: int, instructions: List[Instruction], start_index: int = 1
+) -> Optional[int]:
+    for i in range(start_index, len(instructions)):
+        if instructions[i].offset == target_offset:
+            return i
+    return None
+
+
+def resolved_attrs(instructions: List[Instruction]) -> Tuple[str, int]:
     """ """
     # we can probably speed up using the "tos_str" field.
     resolved = []
@@ -656,7 +822,7 @@ def short_code_repr(code) -> str:
         return f"<code object {code}>"
 
 
-def skip_cache(instructions: list, i: int) -> int:
+def skip_cache(instructions: List[Instruction], i: int) -> int:
     """Python 3.11+ has CACHE instructions.
     Skip over those starting at index i and return
     the index of the first instruction that is not CACHE
@@ -664,7 +830,7 @@ def skip_cache(instructions: list, i: int) -> int:
     such an instruction.
     """
     n = len(instructions)
-    while instructions[i].opname == "CACHE" and i < n:
+    while i < n and instructions[i].opname == "CACHE":
         i += 1
     return i
 
@@ -685,13 +851,14 @@ opcode_extended_fmt_base = {
     "BINARY_OR":             extended_format_BINARY_OR,
     "BINARY_POWER":          extended_format_BINARY_POWER,
     "BINARY_XOR":            extended_format_BINARY_XOR,
+    "BUILD_CONST_KEY_MAP":   extended_format_BUILD_CONST_KEY_MAP,
     "BUILD_LIST":            extended_format_BUILD_LIST,
-    "BUILD_MAP":             extended_format_BUILD_MAP,
     "BUILD_SET":             extended_format_BUILD_SET,
     "BUILD_SLICE":           extended_format_BUILD_SLICE,
     "BUILD_TUPLE":           extended_format_BUILD_TUPLE,
     "CALL_FUNCTION":         extended_format_CALL_FUNCTION,
     "COMPARE_OP":            extended_format_COMPARE_OP,
+    "IMPORT_FROM":           extended_format_IMPORT_FROM,
     "IMPORT_NAME":           extended_format_IMPORT_NAME,
     "INPLACE_ADD":           extended_format_INPLACE_ADD,
     "INPLACE_AND":           extended_format_INPLACE_AND,
@@ -709,10 +876,10 @@ opcode_extended_fmt_base = {
     "LOAD_ATTR":             extended_format_ATTR,
     "LOAD_BUILD_CLASS":      extended_format_LOAD_BUILD_CLASS,
     "MAKE_FUNCTION":         extended_format_MAKE_FUNCTION_10_27,
-    # "LOAD_DEREF":            extended_format_ATTR, # not quite right
     "RAISE_VARARGS":         extended_format_RAISE_VARARGS_older,
     "RETURN_VALUE":          extended_format_RETURN_VALUE,
     "STORE_ATTR":            extended_format_ATTR,
+    "STORE_DEREF":           extended_format_store_op,
     "STORE_FAST":            extended_format_store_op,
     "STORE_GLOBAL":          extended_format_store_op,
     "STORE_NAME":            extended_format_store_op,
